/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/ios/sound.c: iOS audio output interface.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/sound/mixer.h"
#include "src/sysdep.h"
#include "src/sysdep/ios/sound.h"
#include "src/thread.h"
#include "src/time.h"

#include <AudioToolbox/AudioToolbox.h>
#include <AudioUnit/AudioUnit.h>

/*************************************************************************/
/************************* Configuration options *************************/
/*************************************************************************/

/**
 * SOUND_SLICELEN:  Number of samples to send to the hardware for a single
 * call to the render callback.  (The iOS documentation refers to this as
 * the "slice size".)
 */
#define SOUND_SLICELEN  1024

/**
 * SOUND_BUFFERS:  Number of SOUND_SLICELEN-sized buffers to use for
 * buffering audio data.
 */
#define SOUND_BUFFERS  4

/**
 * MIXER_THREAD_PRIORITY:  Thread priority used for the mixer thread.
 */
#define MIXER_THREAD_PRIORITY  5

/*************************************************************************/
/****************** Shared data (internal to iOS code) *******************/
/*************************************************************************/

/**
 * ios_ignore_audio_route_change_delay:  If nonzero, indicates the
 * sys_time_now() timestamp until which audio route changes should be
 * ignored.  This is a workaround for an iOS bug/misfeature (reported to
 * Apple as bug 9677380) which sends outdated route change events to the
 * app when it resumes from suspend.
 */
double ios_ignore_audio_route_change_until;

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Remote I/O audio unit instance for playback. */
static AudioUnit remoteio_instance;

/* Output sampling rate used by the hardware. */
static int output_rate;

/* Flag indicating whether we have an unacknowledged headphone disconnect. */
static uint8_t got_headphone_disconnect;

/*-----------------------------------------------------------------------*/

/* Audio data buffer.  Samples from the software mixer are buffered here
 * before being sent to the hardware. */
static void *output_buffer_mem;
static struct {
    uint8_t full;  // Is this buffer ready to be played?
    void *data;
} output_buffers[SOUND_BUFFERS];

/* Next buffer to send to the hardware (used by audio render callback). */
static unsigned int next_buffer_to_play;

/* Thread ID of mixer thread, and flag used to tell it to stop. */
static int mixer_thread_id;
static uint8_t mixer_thread_stop;

/*-----------------------------------------------------------------------*/

/* Local routine declarations. */

/**
 * audio_render_callback:  Callback called to generate audio data to be
 * sent to the hardware output device.  This function is called in an
 * independent, realtime-priority thread.
 *
 * [Parameters]
 *     userdata: Opaque user data pointer (unused).
 *     flags: Operation flags (unused).
 *     timestamp: Stream timestamp (unused).
 *     bus: Bus index (unused).
 *     num_frames: Number of audio frames (samples) to generate.
 *     buffers: List of buffers into which to store data.
 */
static OSStatus audio_render_callback(void *userdata,
                                      AudioUnitRenderActionFlags *flags,
                                      const AudioTimeStamp *timestamp,
                                      UInt32 bus, UInt32 num_frames,
                                      AudioBufferList *buffers);

/**
 * audio_interruption_callback:  Callback function called by iOS when the
 * audio session is interrupted or resumes from an interruption.
 *
 * [Parameters]
 *     userdata: Opaque user data pointer (unused).
 *     state: Interruption state code.
 */
static void audio_interruption_callback(void *userdata, UInt32 state);

/**
 * audio_route_change_callback:  Callback function called by iOS when the
 * current audio route changes (such as when headphones are plugged in or
 * unplugged).
 *
 * [Parameters]
 *     userdata: Opaque user data pointer (unused).
 *     property: Property ID (always kAudioSessionProperty_AudioRouteChange).
 *     size: Size of value, in bytes.
 *     value: New property value.
 */
static void audio_route_change_callback(void *userdata,
                                        AudioSessionPropertyID property,
                                        UInt32 size, const void *value);

/**
 * mixer_thread:  Thread which buffers the output of the software mixer.
 *
 * [Parameters]
 *     userdata: Opaque user data pointer (unused).
 * [Return value]
 *     0
 */
static int mixer_thread(void *userdata);

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

int sys_sound_init(UNUSED const char *device_name, int num_channels,
                   int *channels_ret)
{
    /* Initialize the iOS audio session. */

    AudioSessionInitialize(NULL, NULL, audio_interruption_callback, NULL);
    const OSStatus result = AudioSessionSetActive(1);
    if (result != kAudioSessionNoError) {
        DLOG("AudioSessionSetActive() failed: %08X", (int)result);
        goto error_return;
    }

    /* Configure the audio session.  (iPhone HIG: Games should use Ambient
     * mode if other audio is playing when the app is started, otherwise
     * Solo Ambient.) */

    uint32_t is_playing = 0;
    UInt32 size = sizeof(is_playing);
    AudioSessionGetProperty(kAudioSessionProperty_OtherAudioIsPlaying,
                            &size, &is_playing);
    const uint32_t category = is_playing
                              ? kAudioSessionCategory_AmbientSound
                              : kAudioSessionCategory_SoloAmbientSound;
    if (0 != AudioSessionSetProperty(kAudioSessionProperty_AudioCategory,
                                     sizeof(category), &category)) {
        DLOG("WARNING: Failed to set audio session category");
    }

    Float64 output_rate_float = 0;
    size = sizeof(output_rate_float);
    AudioSessionGetProperty(kAudioSessionProperty_CurrentHardwareSampleRate,
                            &size, &output_rate_float);
    output_rate = iround(output_rate_float);
    if (output_rate <= 0) {
        DLOG("WARNING: Could not get output rate, assuming 44.1kHz");
        output_rate = 44100;
    }

    Float32 io_buffer_duration = (Float32)SOUND_SLICELEN / (Float32)output_rate;
    if (0 != AudioSessionSetProperty(
                 kAudioSessionProperty_PreferredHardwareIOBufferDuration,
                 sizeof(io_buffer_duration), &io_buffer_duration)) {
        DLOG("Could not set slice size");
        goto error_clear_audio_session;
    }
    size = sizeof(io_buffer_duration);
    if (0 != AudioSessionGetProperty(
                 kAudioSessionProperty_CurrentHardwareIOBufferDuration,
                 &size, &io_buffer_duration)) {
        DLOG("Could not get slice size");
        goto error_clear_audio_session;
    }
    const int slice_len = iroundf(io_buffer_duration * output_rate);
    if (slice_len != SOUND_SLICELEN) {
        DLOG("Wrong slice size: %d, wanted %d", slice_len, SOUND_SLICELEN);
        goto error_clear_audio_session;
    }

    /* Add a callback to catch audio route changes, such as headphones
     * being plugged in or unplugged.  (iPhone HIG: Audio playback should
     * pause when headphones are unplugged or wireless headphones go out
     * of range.)  We add the callback after setting the audio category
     * above to avoid an unnecessary trigger of the callback when the
     * category is set. */

    if (0 != AudioSessionAddPropertyListener(
                 kAudioSessionProperty_AudioRouteChange,
                 audio_route_change_callback, NULL)) {
        DLOG("WARNING: Failed to add audio route change callback");
    }

    /* Set up the software mixer. */

    output_buffer_mem = mem_alloc((4*SOUND_SLICELEN) * SOUND_BUFFERS, 0, 0);
    if (!output_buffer_mem) {
        DLOG("No memory for output buffers (%d bytes)",
             (4*SOUND_SLICELEN) * SOUND_BUFFERS);
        goto error_clear_audio_session;
    }
    for (int i = 0; i < SOUND_BUFFERS; i++) {
        output_buffers[i].full = 0;
        output_buffers[i].data =
            (uint8_t *)output_buffer_mem + (4*SOUND_SLICELEN)*i;
    }

    int retval = sound_mixer_init(num_channels, output_rate);
    if (!retval) {
        DLOG("Mixer initialization failed");
        goto error_free_output_buffer_mem;
    }
    for (int i = 0; i < num_channels; i++) {
        channels_ret[i] = i+1;
    }

    /* Set up a Remote I/O audio unit for playback. */

    AudioComponentDescription remoteio_desc;
    remoteio_desc.componentType         = kAudioUnitType_Output;
    remoteio_desc.componentSubType      = kAudioUnitSubType_RemoteIO;
    remoteio_desc.componentManufacturer = kAudioUnitManufacturer_Apple;
    remoteio_desc.componentFlags        = 0;
    remoteio_desc.componentFlagsMask    = 0;
    AudioComponent remoteio_ref = AudioComponentFindNext(NULL, &remoteio_desc);
    if (0 != AudioComponentInstanceNew(remoteio_ref, &remoteio_instance)) {
        DLOG("Failed to create Remote I/O audio unit");
        goto error_close_mixer;
    }

    AudioStreamBasicDescription format;
    mem_clear(&format, sizeof(format));
    format.mFormatID         = kAudioFormatLinearPCM;
    format.mFormatFlags      = kAudioFormatFlagsCanonical;  // i.e. s16
    format.mBytesPerPacket   = 4;
    format.mBytesPerFrame    = 4;
    format.mFramesPerPacket  = 1;
    format.mBitsPerChannel   = 16;
    format.mChannelsPerFrame = 2;
    format.mSampleRate       = output_rate;
    if (0 != AudioUnitSetProperty(remoteio_instance,
                                  kAudioUnitProperty_StreamFormat,
                                  kAudioUnitScope_Input, 0,
                                  &format, sizeof(format))) {
        DLOG("Failed to set output stream format");
        goto error_destroy_remoteio;
    }

    AURenderCallbackStruct callback;
    callback.inputProc       = &audio_render_callback;
    callback.inputProcRefCon = NULL;
    if (0 != AudioUnitSetProperty(remoteio_instance,
                                  kAudioUnitProperty_SetRenderCallback,
                                  kAudioUnitScope_Input, 0,
                                  &callback, sizeof(callback))) {
        DLOG("Failed to install audio render callback");
        goto error_destroy_remoteio;
    }

    /* Start playback. */

    next_buffer_to_play = 0;
    if (0 != AudioUnitInitialize(remoteio_instance)) {
        DLOG("Failed to initialize audio playback");
        goto error_destroy_remoteio;
    }
    if (0 != AudioOutputUnitStart(remoteio_instance)) {
        DLOG("Failed to start audio playback");
        goto error_uninit_remoteio;
    }

    mixer_thread_stop = 0;
    mixer_thread_id = thread_create_with_priority(
        MIXER_THREAD_PRIORITY, mixer_thread, NULL);
    if (!mixer_thread_id) {
        DLOG("Failed to start mixer thread");
        goto error_stop_remoteio;
    }

    /* All done. */

    return num_channels;


  error_stop_remoteio:
    AudioOutputUnitStop(remoteio_instance);
  error_uninit_remoteio:
    AudioUnitUninitialize(remoteio_instance);
  error_destroy_remoteio:
    AudioComponentInstanceDispose(remoteio_instance);
    remoteio_instance = 0;
  error_close_mixer:
    sound_mixer_cleanup();
  error_free_output_buffer_mem:
    mem_free(output_buffer_mem);
    output_buffer_mem = NULL;
  error_clear_audio_session:
    AudioSessionSetActive(0);
  error_return:
    return 0;
}

/*-----------------------------------------------------------------------*/

int sys_sound_playback_rate(void)
{
    return output_rate;
}

/*-----------------------------------------------------------------------*/

int sys_sound_check_headphone_disconnect(void)
{
    return got_headphone_disconnect;
}

/*-----------------------------------------------------------------------*/

void sys_sound_acknowledge_headphone_disconnect(void)
{
    if (got_headphone_disconnect) {
        if (remoteio_instance) {
            AudioOutputUnitStart(remoteio_instance);
        }
        got_headphone_disconnect = 0;
    }
}

/*-----------------------------------------------------------------------*/

void sys_sound_cleanup(void)
{
    mixer_thread_stop = 1;
    thread_wait(mixer_thread_id);
    mixer_thread_id = 0;

    AudioOutputUnitStop(remoteio_instance);
    AudioUnitUninitialize(remoteio_instance);
    AudioComponentInstanceDispose(remoteio_instance);
    remoteio_instance = 0;

    sound_mixer_cleanup();

    mem_free(output_buffer_mem);
    output_buffer_mem = NULL;

    AudioSessionSetActive(0);
}

/*-----------------------------------------------------------------------*/

/*
 * The remaining sys_sound_*() functions are all redirected to the software
 * mixer functions, which implement the actual channel management.
 */

int sys_sound_setdata(int channel, SoundDecodeCallback decode_func,
                      void *handle, int is_stereo, UNUSED int sample_rate)
{
    return sound_mixer_setdata(channel, decode_func, handle, is_stereo);
}

void sys_sound_setvol(int channel, float volume)
{
    return sound_mixer_setvol(channel, volume);
}

void sys_sound_setpan(int channel, float pan)
{
    return sound_mixer_setpan(channel, pan);
}

void sys_sound_setfade(int channel, float target, float length, int cut)
{
    return sound_mixer_setfade(channel, target, length, cut);
}

void sys_sound_start(int channel)
{
    return sound_mixer_start(channel);
}

void sys_sound_stop(int channel)
{
    return sound_mixer_stop(channel);
}

void sys_sound_reset(int channel)
{
    return sound_mixer_reset(channel);
}

int sys_sound_status(int channel)
{
    return sound_mixer_status(channel);
}

/*************************************************************************/
/*********************** System callback routines ************************/
/*************************************************************************/

static OSStatus audio_render_callback(UNUSED void *userdata,
                                      AudioUnitRenderActionFlags *flags,
                                      UNUSED const AudioTimeStamp *timestamp,
                                      UNUSED UInt32 bus, UInt32 num_frames,
                                      AudioBufferList *buffers)
{
    if (UNLIKELY(num_frames != SOUND_SLICELEN)) {
        static int warned = 0;
        if (!warned) {
            DLOG("Invalid slice length %d", (int)num_frames);
            warned = 1;
        }
        goto fill_with_silence;
    }

    if (UNLIKELY(buffers->mNumberBuffers != 1)) {
        static int warned = 0;
        if (!warned) {
            DLOG("Invalid buffer count %d", (int)buffers->mNumberBuffers);
            warned = 1;
        }
        goto fill_with_silence;
    }

    AudioBuffer *buffer = &buffers->mBuffers[0];
    void *data = buffer->mData;
    if (UNLIKELY(buffer->mNumberChannels != 2)
     || UNLIKELY(buffer->mDataByteSize != 4 * num_frames)) {
        static int warned = 0;
        if (!warned) {
            DLOG("Invalid buffer format: channels=%d size=%d (num_frames=%d)",
                 (int)buffer->mNumberChannels, (int)buffer->mDataByteSize,
                 (int)num_frames);
            warned = 1;
        }
        goto fill_with_silence;
    }

    BARRIER();
    if (!output_buffers[next_buffer_to_play].full) {
        goto fill_with_silence;
    }

    memcpy(data, output_buffers[next_buffer_to_play].data,
           buffer->mDataByteSize);
    output_buffers[next_buffer_to_play].full = 0;
    next_buffer_to_play = (next_buffer_to_play + 1) % SOUND_BUFFERS;
    return 0;

  fill_with_silence:;
    for (unsigned int i = 0; i < buffers->mNumberBuffers; i++) {
        mem_clear(buffers->mBuffers[i].mData,
                  buffers->mBuffers[i].mDataByteSize);
    }
    *flags |= kAudioUnitRenderAction_OutputIsSilence;
    return 0;
}

/*-----------------------------------------------------------------------*/

static void audio_interruption_callback(UNUSED void *userdata, UInt32 state)
{
    if (state == kAudioSessionBeginInterruption) {
        if (remoteio_instance) {
            AudioOutputUnitStop(remoteio_instance);
        }
    } else if (state == kAudioSessionEndInterruption) {
        if (remoteio_instance) {
            AudioOutputUnitStart(remoteio_instance);
        }
    } else {
        DLOG("Unknown state 0x%08X", (int)state);
    }
}

/*-----------------------------------------------------------------------*/

static void audio_route_change_callback(UNUSED void *userdata,
                                        AudioSessionPropertyID property,
                                        UNUSED UInt32 size, const void *value)
{
    if (ios_ignore_audio_route_change_until != 0) {
        if (sys_time_now() < ios_ignore_audio_route_change_until) {
            return;
        } else {
            ios_ignore_audio_route_change_until = 0;
        }
    }

    if (property != kAudioSessionProperty_AudioRouteChange) {
        DLOG("Got unexpected property ID 0x%08X", (int)property);
        return;
    }

    CFDictionaryRef dict = (CFDictionaryRef)value;
    CFNumberRef reason_ref = CFDictionaryGetValue(dict, CFSTR(kAudioSession_AudioRouteChangeKey_Reason));
    int32_t reason;
    CFNumberGetValue(reason_ref, kCFNumberSInt32Type, &reason);

    if (reason == kAudioSessionRouteChangeReason_OldDeviceUnavailable) {
        /* The built-in speaker can never become unavailable, so this must
         * be the result of headphones or some other such device being
         * disconnected. */
        if (!got_headphone_disconnect) {
            if (remoteio_instance) {
                AudioOutputUnitStop(remoteio_instance);
            }
            got_headphone_disconnect = 1;
        }
    }
}

/*************************************************************************/
/************************* Other local routines **************************/
/*************************************************************************/

static int mixer_thread(UNUSED void *userdata)
{
    const float slice_time = (float)SOUND_SLICELEN / (float)output_rate;

    unsigned int next_buffer_to_fill = 0;
    do {
        BARRIER();
        /*
         * In theory, it's enough to check whether the next buffer to fill
         * is empty or not.  However, if we ever happen to get out of sync
         * with the hardware output thread, we'd end up playing buffers in
         * a different order from how they were generated:
         *
         *     to_fill  to_play  Buffer 0  Buffer 1  Action taken
         *     -------  -------  --------  --------  -------------
         *        0        1       empty     empty   Fill 0 with A
         *        1        1         A       empty   Fill 1 with B
         *        0        1         A         B     Play B
         *        0        0         A       empty   Play A
         *        0        1       empty     empty   Fill 0 with C
         *        1        1         C       empty   Fill 1 with D
         *        0        1         C         D     Play D
         *        0        0         C       empty   Play C
         *        0        1       empty     empty   ...
         *
         * So to be safe, if the next buffer to fill is already full, we
         * look for the next non-empty buffer and fill it instead of just
         * waiting for the expected buffer to become empty:
         *
         *     to_fill  to_play  Buffer 0  Buffer 1  Action taken
         *     -------  -------  --------  --------  -------------
         *        0        1       empty     empty   Fill 0 with A
         *        1        1         A       empty   Fill 1 with B
         *        0        1         A         B     Play B
         *        0        0         A       empty   Fill 1 with C
         *        0        0         A         C     Play A
         *        0        1       empty       C     Fill 0 with D
         *        1        1         D         C     Play C
         *        1        0         D       empty   Fill 1 with E
         *        0        0         D         E     Play D
         *        0        1       empty       E     ...
         *
         * This relies on the assumption that the mixer will generally
         * run faster than the playback thread -- more specifically, that
         * this loop will execute more frequently than the audio render
         * callback is called by the system.
         */
        const unsigned int old_to_fill = next_buffer_to_fill;
        if (output_buffers[next_buffer_to_fill].full) {
            do {
                next_buffer_to_fill = (next_buffer_to_fill + 1) % SOUND_BUFFERS;
            } while (next_buffer_to_fill != old_to_fill
                     && output_buffers[next_buffer_to_fill].full);
        }
        if (!output_buffers[next_buffer_to_fill].full) {
            sound_mixer_get_pcm(output_buffers[next_buffer_to_fill].data,
                                SOUND_SLICELEN);
            output_buffers[next_buffer_to_fill].full = 1;
            next_buffer_to_fill = (next_buffer_to_fill + 1) % SOUND_BUFFERS;
        } else {
            time_delay(slice_time/2);  // Must be < slice_time (see above).
        }
    } while (!mixer_thread_stop);
    return 0;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
