/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/opengl/shader.c: Shader management functionality for
 * OpenGL-based systems.
 */

#include "src/base.h"
#include "src/graphics.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/opengl/internal.h"
#include "src/sysdep/opengl/opengl.h"
#include "src/sysdep/opengl/shader-table.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Value used in ShaderInfo.program meaning "shader creation failed".  This
 * is used to avoid trying to recreate the same shader over and over when
 * we already know it won't work.  We use (GLuint)-1 on the assumption
 * that the GL numbers program IDs from 1 up (as most seem to do), but
 * create_shader() checks for a collision with this ID and explicitly
 * leaks the program object and tries again if a collision occurs. */
#define INVALID_PROGRAM_ID  ((GLuint)-1)

/* Callback functions for generating shader source code and key values. */
static ShaderSourceCallback *shader_source_callback;
static ShaderKeyCallback *shader_key_callback;

/* Shader program currently in use. */
static ShaderInfo *current_shader = NULL;

/*-----------------------------------------------------------------------*/

/* Constants for the "type" parameter to get_shader_string(). */

typedef enum ShaderStringType ShaderStringType;
enum ShaderStringType {
    GET_SHADER_SOURCE,  // The shader's source code.
    GET_SHADER_INFO,    // The shader's information log.
    GET_PROGRAM_INFO,   // The shader program's information log.
};

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * create_shader:  Create a new shader for the given vertex type.
 *
 * [Parameters]
 *     primitive_type: Primitive type (GL_TRIANGLES, etc.).
 *     position_count: Number of position elements (2-4).
 *     texcoord_count: Number of texture coordinate elements (0 or 2).
 *     texcolor_type: Texture color type (GL_RGBA or GL_ALPHA, or 0 for
 *         OpenGL  external textures); ignored if texturing is disabled.
 *     tex_offset: True for an external texture offset passed as a uniform
 *         parameter, false for none.
 *     color_count: Number of color elements (0 or 4).
 *     color_uniform: True for an external fixed color passed as a uniform
 *         parameter, false for none.
 *     fog: True for linear fog, false for no fog.
 *     alpha_test: True for alpha testing (discard pixels with alpha less
 *         than the reference value), false for no alpha testing.
 *     alpha_comparison: Alpha test comparison type (GL_GEQUAL, etc.).
 * [Return value]
 *     OpenGL program object, or zero on error.
 */
static GLuint create_shader(
    GLenum primitive_type, int position_count, int texcoord_count,
    GLenum texcolor_type, int tex_offset, int color_count,
    int color_uniform, int fog, int alpha_test, GLenum alpha_comparison);

/**
 * generate_shader_source:  Generate and load shader source code compatible
 * with GLSL 1.10 or GLSL ES 1.0 (depending on whether SIL_OPENGL_ES is
 * defined) for the given vertex type.
 *
 * [Parameters]
 *     vertex_shader: OpenGL vertex shader object.
 *     fragment_shader: OpenGL fragment shader object.
 *     primitive_type: Primitive type (GL_TRIANGLES, etc.).
 *     position_count: Number of position elements (2-4).
 *     texcoord_count: Number of texture coordinate elements (0 or 2).
 *     texcolor_type: Texture color type (GL_RGBA or GL_ALPHA, or 0 for
 *         OpenGL  external textures); ignored if texturing is disabled.
 *     tex_offset: True for an external texture offset passed as a uniform
 *         parameter, false for none.
 *     color_count: Number of color elements (0 or 4).
 *     color_uniform: True for an external fixed color passed as a uniform
 *         parameter, false for none.
 *     fog: True for linear fog, false for no fog.
 *     alpha_test: True for alpha testing (discard pixels with alpha less
 *         than the reference value), false for no alpha testing.
 *     alpha_comparison: Alpha test comparison type (GL_GEQUAL, etc.).
 * [Return value]
 *     True on success, false on error.
 */
static int generate_shader_source(
    GLuint vertex_shader, GLuint fragment_shader,
    GLenum primitive_type, int position_count, int texcoord_count,
    GLenum texcolor_type, int texoffset_uniform, int color_count,
    int color_uniform, int fog, int alpha_test, GLenum alpha_comparison);

/**
 * generate_shader_key:  Return a unique 32-bit key for a shader with the
 * given parameters.
 *
 * [Parameters]
 *     As for generate_shader_source(), but does not take vertex_shader or
 *     fragment_shader.
 * [Return value]
 *     Unique 32-bit shader key.
 */
static uint32_t generate_shader_key(
    GLenum primitive_type, int position_count, int texcoord_count,
    GLenum texcolor_type, int texoffset_uniform, int color_count,
    int color_uniform, int fog, int alpha_test, GLenum alpha_comparison);

#ifdef DEBUG
/**
 * get_shader_string:  Retrieve string data for a shader or program object.
 * Only available when building in debug mode.
 *
 * [Parameters]
 *     type: Type of data to retrieve (GET_SHADER_SOURCE, etc.).
 *     object: Shader or program object.
 * [Return value]
 *     String data in a mem_alloc()ed buffer, or NULL on error.
 */
static char *get_shader_string(ShaderStringType type, GLuint object);
#endif

/*************************************************************************/
/*********************** Exported utility routines ***********************/
/*************************************************************************/

void opengl_set_shader_callbacks(
    ShaderSourceCallback *source_callback, ShaderKeyCallback *key_callback,
    int hash_table_size, int dynamic_resize)
{
    PRECOND(source_callback == NULL || key_callback != NULL,
            source_callback = NULL);
    PRECOND(source_callback == NULL || hash_table_size > 0 || dynamic_resize,
            source_callback = NULL);

    if (source_callback) {
        shader_source_callback = source_callback;
        shader_key_callback = key_callback;
        shader_table_init(hash_table_size, dynamic_resize);
    } else {
        shader_source_callback = generate_shader_source;
        shader_key_callback = generate_shader_key;
        shader_table_init(2 *  // primitive_type==GL_POINTS ? 1 : 0
                          3 *  // lenof(position_counts)
                          4 *  // 1 + (lenof(texcoord_counts)
                               //      * lenof(texcolor_types))
                          2 *  // tex_offset ? 1 : 0
                          2 *  // lenof(color_counts)
                          2 *  // color_uniform ? 1 : 0
                          2 *  // fog ? 1 : 0
                          5,   // 1 + lenof(alpha_comparison_types)
                          0);  // All shaders covered, so no need to resize.
    }
}

/*************************************************************************/
/******************* Library-internal utility routines *******************/
/*************************************************************************/

int opengl_select_shader(const SysPrimitive *primitive,
                         int tex_offset, int color_uniform,
                         int fog, int alpha_test, GLenum alpha_comparison)
{
    /* Extract data about the current primitive.  If there's no texture
     * currently applied, ignore any texture coordinate in the vertex data. */

    int position_count, texcoord_count, color_count;
    GLenum texcolor_type;

    position_count = primitive->position_size;
    if (opengl_texture_is_framebuffer || opengl_current_texture != NULL) {
        texcoord_count = primitive->texcoord_size;
    } else {
        texcoord_count = 0;
    }
    if (opengl_texture_is_framebuffer) {
        texcolor_type = GL_RGBA;
    } else if (opengl_current_texture != NULL) {
        texcolor_type =
            (opengl_current_texture->color_type==TEXCOLOR_A ? GL_ALPHA :
             opengl_current_texture->color_type==TEXCOLOR_EXTERNAL ? 0 :
             GL_RGBA);
    } else {
        texcolor_type = GL_RGBA;
    }
    color_count = primitive->color_size;

    /* Look up the ShaderInfo structure for this vertex data type. */

    const uint32_t key = (*shader_key_callback)(
        primitive->type, position_count, texcoord_count, texcolor_type,
        tex_offset, color_count, color_uniform, fog, alpha_test,
        alpha_comparison);
    if (key == INVALID_SHADER_KEY) {
        return -1;
    }
    int invalidate = 1; // Play it safe and null out current_shader by default.
    ShaderInfo * const shader_info = shader_table_lookup(key, &invalidate);
    if (invalidate) {
        current_shader = NULL;  // Old pointer is no longer valid.
    }
    if (UNLIKELY(!shader_info)) {
        DLOG("Failed to look up shader for key %u", key);
        return -1;
    }

    /* Create the shader program if necessary. */

    if (!shader_info->program) {
        const GLuint new_program = create_shader(
            primitive->type, position_count, texcoord_count, texcolor_type,
            tex_offset, color_count, color_uniform, fog, alpha_test,
            alpha_comparison);
        if (UNLIKELY(!new_program)) {
            DLOG("Failed to create shader for position=%d texcoord=%d/%d"
                 " texcolor=%d color=%d/%d", position_count, texcoord_count,
                 tex_offset, texcolor_type, color_count, color_uniform);
            shader_info->program = INVALID_PROGRAM_ID;
            return -1;
        }
        shader_info->program = new_program;
        shader_info->uniforms[UNIFORM_TRANSFORM] =
            glGetUniformLocation(new_program, "transform");
        shader_info->uniforms[UNIFORM_TEXTURE] =
            glGetUniformLocation(new_program, "texture");
        shader_info->uniforms[UNIFORM_TEX_OFFSET] =
            glGetUniformLocation(new_program, "tex_offset");
        shader_info->uniforms[UNIFORM_FIXED_COLOR] =
            glGetUniformLocation(new_program, "fixed_color");
        shader_info->uniforms[UNIFORM_FOG_PARAMS] =
            glGetUniformLocation(new_program, "fog_params");
        shader_info->uniforms[UNIFORM_FOG_COLOR] =
            glGetUniformLocation(new_program, "fog_color");
        shader_info->uniforms[UNIFORM_ALPHA_REF] =
            glGetUniformLocation(new_program, "alpha_ref");
        shader_info->uniforms[UNIFORM_POINT_SIZE] =
            glGetUniformLocation(new_program, "point_size");
    }

    /* Activate the shader program if it's not already active. */

    int changed = 0;
    if (shader_info != current_shader) {
        if (shader_info->program == INVALID_PROGRAM_ID) {
            return -1;
        }
        current_shader = shader_info;
        changed = 1;
        glUseProgram(current_shader->program);
    }

    return changed;
}

/*-----------------------------------------------------------------------*/

void opengl_set_uniform_int(int uniform, int value)
{
    PRECOND(uniform >= 0 && uniform < UNIFORM__NUM, return);
    if (current_shader) {
        glUniform1i(current_shader->uniforms[uniform], value);
    }
}

/*-----------------------------------------------------------------------*/

void opengl_set_uniform_float(int uniform, float value)
{
    PRECOND(uniform >= 0 && uniform < UNIFORM__NUM, return);
    if (current_shader) {
        glUniform1f(current_shader->uniforms[uniform], value);
    }
}

/*-----------------------------------------------------------------------*/

void opengl_set_uniform_vec2(int uniform, const Vector2f *value)
{
    PRECOND(uniform >= 0 && uniform < UNIFORM__NUM, return);
    if (current_shader) {
        glUniform2f(current_shader->uniforms[uniform], value->x, value->y);
    }
}

/*-----------------------------------------------------------------------*/

void opengl_set_uniform_vec4(int uniform, const Vector4f *value)
{
    PRECOND(uniform >= 0 && uniform < UNIFORM__NUM, return);
    if (current_shader) {
        glUniform4fv(current_shader->uniforms[uniform], 1, &value->x);
    }
}

/*-----------------------------------------------------------------------*/

void opengl_set_uniform_mat4(int uniform, const Matrix4f *value)
{
    PRECOND(uniform >= 0 && uniform < UNIFORM__NUM, return);
    if (current_shader) {
        int transpose;
#ifdef SIL_OPENGL_TRANSPOSE_MATRIX
        transpose = 1;
#else
        transpose = 0;
#endif
#ifdef SIL_OPENGL_ES
        /* transpose == TRUE is unsupported on OpenGL ES. */
        Matrix4f value_t;
        if (transpose) {
            value_t._11 = value->_11;
            value_t._12 = value->_21;
            value_t._13 = value->_31;
            value_t._14 = value->_41;
            value_t._21 = value->_12;
            value_t._22 = value->_22;
            value_t._23 = value->_32;
            value_t._24 = value->_42;
            value_t._31 = value->_13;
            value_t._32 = value->_23;
            value_t._33 = value->_33;
            value_t._34 = value->_43;
            value_t._41 = value->_14;
            value_t._42 = value->_24;
            value_t._43 = value->_34;
            value_t._44 = value->_44;
            value = &value_t;
            transpose = 0;
        }
#endif
        glUniformMatrix4fv(current_shader->uniforms[uniform], 1, transpose,
                           &value->_11);
    }
}

/*-----------------------------------------------------------------------*/

void opengl_shader_clear_all(void)
{
    shader_table_clear();
    current_shader = NULL;
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static GLuint create_shader(
    GLenum primitive_type, int position_count, int texcoord_count,
    GLenum texcolor_type, int tex_offset, int color_count,
    int color_uniform, int fog, int alpha_test, GLenum alpha_comparison)
{
    GLint ok, error;

    opengl_clear_error();

    /* Create the OpenGL program and shader resources. */

    const GLuint program = glCreateProgram();
    if (UNLIKELY(!program)) {
        DLOG("Failed to create new shader program");
        goto error_return;
    }
    if (UNLIKELY(program == INVALID_PROGRAM_ID)) {
        DLOG("Shader program collided with INVALID_PROGRAM_ID (%u),"
             " leaking it and generating another", INVALID_PROGRAM_ID);
        return create_shader(primitive_type, position_count, texcoord_count,
                             texcolor_type, tex_offset, color_count,
                             color_uniform, fog, alpha_test, alpha_comparison);
    }
    const GLuint vertex_shader = glCreateShader(GL_VERTEX_SHADER);
    if (UNLIKELY(!vertex_shader)) {
        DLOG("Failed to create new vertex shader");
        goto error_delete_program;
    }
    const GLuint fragment_shader = glCreateShader(GL_FRAGMENT_SHADER);
    if (UNLIKELY(!fragment_shader)) {
        DLOG("Failed to create new vertex shader");
        goto error_delete_vertex_shader;
    }
    ASSERT(glGetError() == GL_NO_ERROR, goto error_delete_fragment_shader);

    /* Bind attribute indices to their respective names. */

    glBindAttribLocation(program, VERTEX_ATTRIB_POSITION, "position");
    glBindAttribLocation(program, VERTEX_ATTRIB_TEXCOORD, "texcoord");
    glBindAttribLocation(program, VERTEX_ATTRIB_COLOR,    "color");

#if defined(SIL_OPENGL_SHADER_VERSION) && SIL_OPENGL_SHADER_VERSION >= 130
    glBindFragDataLocation(program, 0, "frag_color");
#endif

    error = glGetError();
    if (UNLIKELY(error != GL_NO_ERROR)) {
        DLOG("Failed to bind shader variables: 0x%04X", error);
        goto error_delete_fragment_shader;
    }

    /* Generate source code for the shaders. */

    ok = (*shader_source_callback)(
        vertex_shader, fragment_shader,
        primitive_type, position_count, texcoord_count, texcolor_type,
        tex_offset, color_count, color_uniform, fog, alpha_test,
        alpha_comparison);
    if (UNLIKELY(!ok)) {
        goto error_delete_fragment_shader;
    }

    /* Compile the shaders. */

    glCompileShader(vertex_shader);
    glGetShaderiv(vertex_shader, GL_COMPILE_STATUS, &ok);
    if (UNLIKELY(!ok)) {
#ifdef DEBUG
        char *info = get_shader_string(GET_SHADER_INFO, vertex_shader);
        char *source = get_shader_string(GET_SHADER_SOURCE, vertex_shader);
        DLOG("Failed to compile vertex shader!  Log follows:\n%s\n"
             "Shader source code follows:\n"
             "================\n%s================", info, source);
        mem_free(info);
        mem_free(source);
#endif
        goto error_delete_fragment_shader;
    }
    ASSERT(glGetError() == GL_NO_ERROR, goto error_delete_fragment_shader);

    glCompileShader(fragment_shader);
    glGetShaderiv(fragment_shader, GL_COMPILE_STATUS, &ok);
    if (UNLIKELY(!ok)) {
#ifdef DEBUG
        char *info = get_shader_string(GET_SHADER_INFO, fragment_shader);
        char *source = get_shader_string(GET_SHADER_SOURCE, fragment_shader);
        DLOG("Failed to compile fragment shader!  Log follows:\n%s\n"
             "Shader source code follows:\n"
             "================\n%s================", info, source);
        mem_free(info);
        mem_free(source);
#endif
        goto error_delete_fragment_shader;
    }
    ASSERT(glGetError() == GL_NO_ERROR, goto error_delete_fragment_shader);

    /* Link the vertex and shader fragment together to create the shader
     * program. */

    glAttachShader(program, vertex_shader);
    glAttachShader(program, fragment_shader);
    glLinkProgram(program);
    glGetProgramiv(program, GL_LINK_STATUS, &ok);
    if (UNLIKELY(!ok)) {
#ifdef DEBUG
        char *info = get_shader_string(GET_PROGRAM_INFO, program);
        DLOG("Failed to link shader program!  Log follows:\n%s", info);
        mem_free(info);
#endif
        goto error_delete_fragment_shader;
    }
    ASSERT(glGetError() == GL_NO_ERROR, goto error_delete_fragment_shader);

    /* Success!  Dump the shader source and logs if requested, and return
     * the new program. */

#if defined(DEBUG) && defined(SIL_OPENGL_DUMP_SHADERS)
    char *vert_source = get_shader_string(GET_SHADER_SOURCE, vertex_shader);
    char *vert_info = get_shader_string(GET_SHADER_INFO, vertex_shader);
    char *frag_source = get_shader_string(GET_SHADER_SOURCE,fragment_shader);
    char *frag_info = get_shader_string(GET_SHADER_INFO, fragment_shader);
    char *link_info = get_shader_string(GET_PROGRAM_INFO, program);
    DLOG("\n"
         "******** BEGIN SHADER DUMP for position=%d texcoord=%d%s color=%d%s ********\n"
         "\n"
         "Vertex shader source:\n"
         "================\n"
         "%s"
         "================\n"
         "\n"
         "Vertex shader compile log:\n"
         "================\n"
         "%s"
         "================\n"
         "\n"
         "Fragment shader source:\n"
         "================\n"
         "%s"
         "================\n"
         "\n"
         "Fragment shader compile log:\n"
         "================\n"
         "%s"
         "================\n"
         "\n"
         "Program link log:\n"
         "================\n"
         "%s"
         "================\n"
         "\n"
         "********* END SHADER DUMP for position=%d texcoord=%d%s color=%d%s *********\n",
         position_count,
             texcoord_count, tex_offset ? "+offset" : "",
             color_count, color_uniform ? "*fixed" : "",
         vert_source ? vert_source : "",
         vert_info ? vert_info : "",
         frag_source ? frag_source : "",
         frag_info ? frag_info : "",
         link_info ? link_info : "",
         position_count,
             texcoord_count, tex_offset ? "+offset" : "",
             color_count, color_uniform ? "*fixed" : "");
    mem_free(vert_source);
    mem_free(vert_info);
    mem_free(frag_source);
    mem_free(frag_info);
    mem_free(link_info);
#endif

    return program;

    /* Avoid resource leaks on error. */

  error_delete_fragment_shader:
    opengl_delete_shader(fragment_shader);
  error_delete_vertex_shader:
    opengl_delete_shader(vertex_shader);
  error_delete_program:
    opengl_delete_program(program);
  error_return:
    return 0;
}

/*-----------------------------------------------------------------------*/

static int generate_shader_source(
    GLuint vertex_shader, GLuint fragment_shader,
    GLenum primitive_type, int position_count, int texcoord_count,
    GLenum texcolor_type, int tex_offset, int color_count,
    int color_uniform, int fog, int alpha_test, GLenum alpha_comparison)
{
    const char *lines[100];
    int num_lines;
    #define ADD_LINE(s) \
        (num_lines < lenof(lines) && (lines[num_lines++] = (s)))

    #ifdef SIL_OPENGL_SHADER_VERSION
    # define SIL_OPENGL_SHADER_VERSION_STRING _STR1(SIL_OPENGL_SHADER_VERSION)
    # define _STR1(x) _STR2(x)
    # define _STR2(x) #x
    #endif

    #if defined(SIL_OPENGL_SHADER_VERSION) && SIL_OPENGL_SHADER_VERSION >= 130
    # define VERTEX_IN    "in"
    # define VERTEX_OUT   "out"
    # define FRAGMENT_IN  "in"
    # define GL_FRAGCOLOR "frag_color"
    # define TEXTURE2D    "texture"
    #else
    # define VERTEX_IN    "attribute"
    # define VERTEX_OUT   "varying"
    # define FRAGMENT_IN  "varying"
    # define GL_FRAGCOLOR "gl_FragColor"
    # define TEXTURE2D    "texture2D"
    #endif

    #ifdef SIL_OPENGL_ES
    # define HIGHP   "highp"
    # define MEDIUMP "mediump"
    # define LOWP    "lowp"
    # define TEXP    "texp"  // "highp" if available, else "mediump".
    #else
    # define HIGHP   ""
    # define MEDIUMP ""
    # define LOWP    ""
    # define TEXP    ""
    #endif


    /**** Vertex shader. ****/

    num_lines = 0;

#ifdef SIL_OPENGL_SHADER_VERSION
    ASSERT(ADD_LINE("#version "SIL_OPENGL_SHADER_VERSION_STRING"\n"));
#endif

    /* Some reference materials recommend the use of mediump for texture
     * coordinates, but the precision of mediump is only 1 part in 1024,
     * which is insufficient when rendering textures from texture atlases
     * larger than 1024x1024 texels.  We thus default to using highp for
     * texture coordinates, if available in the fragment language.  To
     * avoid compilation failures when highp is not available, we define
     * a precision "texp" which is either highp or mediump depending on
     * the GL's fragment shader capabilities. */
#ifdef SIL_OPENGL_ES
    ASSERT(ADD_LINE("#ifdef GL_FRAGMENT_PRECISION_HIGH\n"));
    ASSERT(ADD_LINE("#define texp highp\n"));
    ASSERT(ADD_LINE("#else\n"));
    ASSERT(ADD_LINE("#define texp mediump\n"));
    ASSERT(ADD_LINE("#endif\n"));
#endif

    ASSERT(ADD_LINE("uniform "HIGHP" mat4 transform;\n"));
    if (tex_offset) {
        ASSERT(ADD_LINE("uniform "TEXP" vec2 tex_offset;\n"));
    }
    if (color_uniform && color_count) {
        ASSERT(ADD_LINE("uniform "LOWP" vec4 fixed_color;\n"));
    }
    if (primitive_type == GL_POINTS) {
        ASSERT(ADD_LINE("uniform "MEDIUMP" float point_size;\n"));
    }
    switch (position_count) {
      case 2:
        ASSERT(ADD_LINE(VERTEX_IN" "HIGHP" vec2 position;\n"));
        break;
      case 3:
        ASSERT(ADD_LINE(VERTEX_IN" "HIGHP" vec3 position;\n"));
        break;
      case 4:
        ASSERT(ADD_LINE(VERTEX_IN" "HIGHP" vec4 position;\n"));
        break;
    }
    switch (texcoord_count) {
      case 2:
        ASSERT(ADD_LINE(VERTEX_IN" "TEXP" vec2 texcoord;\n"));
        ASSERT(ADD_LINE(VERTEX_OUT" "TEXP" vec2 texcoord_varying;\n"));
        break;
    }
    switch (color_count) {
      case 4:
        ASSERT(ADD_LINE(VERTEX_IN" "LOWP" vec4 color;\n"));
        ASSERT(ADD_LINE(VERTEX_OUT" "LOWP" vec4 color_varying;\n"));
        break;
    }
    if (fog) {
        ASSERT(ADD_LINE("uniform "HIGHP" vec4 fog_params;\n"));
        ASSERT(ADD_LINE(VERTEX_OUT" "MEDIUMP" float fog_varying;\n"));
    }

    ASSERT(ADD_LINE("void main() {\n"));

    switch (position_count) {
#ifdef SIL_OPENGL_TRANSPOSE_MATRIX
      case 2:
        ASSERT(ADD_LINE("    gl_Position.x = dot(transform[0], vec4(position, 0.0, 1.0));\n"));
        ASSERT(ADD_LINE("    gl_Position.y = dot(transform[1], vec4(position, 0.0, 1.0));\n"));
        ASSERT(ADD_LINE("    gl_Position.z = dot(transform[2], vec4(position, 0.0, 1.0));\n"));
        ASSERT(ADD_LINE("    gl_Position.w = dot(transform[3], vec4(position, 0.0, 1.0));\n"));
        break;
      case 3:
        ASSERT(ADD_LINE("    gl_Position.x = dot(transform[0], vec4(position, 1.0));\n"));
        ASSERT(ADD_LINE("    gl_Position.y = dot(transform[1], vec4(position, 1.0));\n"));
        ASSERT(ADD_LINE("    gl_Position.z = dot(transform[2], vec4(position, 1.0));\n"));
        ASSERT(ADD_LINE("    gl_Position.w = dot(transform[3], vec4(position, 1.0));\n"));
        break;
      case 4:
        ASSERT(ADD_LINE("    gl_Position.x = dot(transform[0], position);\n"));
        ASSERT(ADD_LINE("    gl_Position.y = dot(transform[1], position);\n"));
        ASSERT(ADD_LINE("    gl_Position.z = dot(transform[2], position);\n"));
        ASSERT(ADD_LINE("    gl_Position.w = dot(transform[3], position);\n"));
        break;
#else  // !SIL_OPENGL_TRANSPOSE_MATRIX
      case 2:
        ASSERT(ADD_LINE("    gl_Position = transform * vec4(position, 0.0, 1.0);\n"));
        break;
      case 3:
        ASSERT(ADD_LINE("    gl_Position = transform * vec4(position, 1.0);\n"));
        break;
      case 4:
        ASSERT(ADD_LINE("    gl_Position = transform * position;\n"));
        break;
#endif
    }

    switch (texcoord_count) {
      case 2:
        if (tex_offset) {
            ASSERT(ADD_LINE("    texcoord_varying = texcoord + tex_offset;\n"));
        } else {
            ASSERT(ADD_LINE("    texcoord_varying = texcoord;\n"));
        }
        break;
    }

    switch (color_count) {
      case 4:
        if (color_uniform) {
            ASSERT(ADD_LINE("    color_varying = color * fixed_color;\n"));
        } else {
            ASSERT(ADD_LINE("    color_varying = color;\n"));
        }
        break;
    }

    if (fog) {
        if (position_count >= 3) {
            ASSERT(ADD_LINE("    "HIGHP" float distance = position.z * fog_params.x + fog_params.y;\n"));
        } else {
            ASSERT(ADD_LINE("    "HIGHP" float distance = fog_params.y;\n"));
        }
        ASSERT(ADD_LINE("    fog_varying = (abs(distance) - fog_params.z) * fog_params.w;\n"));
    }

    if (primitive_type == GL_POINTS) {
        ASSERT(ADD_LINE("    gl_PointSize = point_size;\n"));
    }

    ASSERT(ADD_LINE("}\n"));

    glShaderSource(vertex_shader, num_lines, lines, NULL);


    /**** Fragment shader. ****/

    num_lines = 0;

#ifdef SIL_OPENGL_SHADER_VERSION
    ASSERT(ADD_LINE("#version "SIL_OPENGL_SHADER_VERSION_STRING"\n"));
#endif

#ifdef SIL_OPENGL_ES
    ASSERT(ADD_LINE("#ifdef GL_FRAGMENT_PRECISION_HIGH\n"));
    ASSERT(ADD_LINE("#define texp highp\n"));
    ASSERT(ADD_LINE("#else\n"));
    ASSERT(ADD_LINE("#define texp mediump\n"));
    ASSERT(ADD_LINE("#endif\n"));
    if (texcolor_type == 0) {
        ASSERT(ADD_LINE("#extension GL_OES_EGL_image_external : require\n"));
    }
#endif
    if (color_uniform && !color_count) {
        ASSERT(ADD_LINE("uniform "LOWP" vec4 fixed_color;\n"));
    }
    if (texcoord_count > 0) {
        if (texcolor_type == 0) {
#ifdef SIL_OPENGL_ES
            ASSERT(ADD_LINE("uniform "LOWP" samplerExternalOES tex;\n"));
#else
            DLOG("External textures not supported in non-ES OpenGL");
            return 0;
#endif
        } else {
            ASSERT(ADD_LINE("uniform "LOWP" sampler2D tex;\n"));
        }
        ASSERT(ADD_LINE(FRAGMENT_IN" "TEXP" vec2 texcoord_varying;\n"));
    }
    if (color_count > 0) {
        ASSERT(ADD_LINE(FRAGMENT_IN" "LOWP" vec4 color_varying;\n"));
    }
    if (fog) {
        ASSERT(ADD_LINE(FRAGMENT_IN" "MEDIUMP" float fog_varying;\n"));
        ASSERT(ADD_LINE("uniform "LOWP" vec4 fog_color;\n"));
    }
    if (alpha_test) {
        ASSERT(ADD_LINE("uniform "LOWP" float alpha_ref;\n"));
    }

#if defined(SIL_OPENGL_SHADER_VERSION) && SIL_OPENGL_SHADER_VERSION >= 130
    ASSERT(ADD_LINE("out "LOWP" vec4 frag_color;\n"));
#endif

    ASSERT(ADD_LINE("void main() {\n"));

    ASSERT(ADD_LINE("    "LOWP" vec4 color =\n"));
    if (texcoord_count > 0) {
        if (color_count > 0) {
            if (texcolor_type == GL_ALPHA) {
                ASSERT(ADD_LINE("        vec4(color_varying.rgb, color_varying.a * "TEXTURE2D"(tex, texcoord_varying).a);\n"));
            } else {
                ASSERT(ADD_LINE("        "TEXTURE2D"(tex, texcoord_varying) * color_varying;\n"));
            }
        } else if (color_uniform) {
            if (texcolor_type == GL_ALPHA) {
                ASSERT(ADD_LINE("        vec4(fixed_color.rgb, fixed_color.a * "TEXTURE2D"(tex, texcoord_varying).a);\n"));
            } else {
                ASSERT(ADD_LINE("        "TEXTURE2D"(tex, texcoord_varying) * fixed_color;\n"));
            }
        } else {
            if (texcolor_type == GL_ALPHA) {
                ASSERT(ADD_LINE("        vec4(1, 1, 1, "TEXTURE2D"(tex, texcoord_varying).a);\n"));
            } else {
                ASSERT(ADD_LINE("        "TEXTURE2D"(tex, texcoord_varying);\n"));
            }
        }
    } else {
        if (color_count > 0) {
            ASSERT(ADD_LINE("        color_varying;\n"));
        } else if (color_uniform) {
            ASSERT(ADD_LINE("        fixed_color;\n"));
        } else {
            ASSERT(ADD_LINE("        vec4(1.0, 1.0, 1.0, 1.0);\n"));
        }
    }

    if (alpha_test) {
        switch (alpha_comparison) {
          case GL_LESS:
            ASSERT(ADD_LINE("    if (color.a >= alpha_ref) discard;\n"));
            break;
          case GL_LEQUAL:
            ASSERT(ADD_LINE("    if (color.a > alpha_ref) discard;\n"));
            break;
          case GL_GEQUAL:
            ASSERT(ADD_LINE("    if (color.a < alpha_ref) discard;\n"));
            break;
          case GL_GREATER:
            ASSERT(ADD_LINE("    if (color.a <= alpha_ref) discard;\n"));
            break;
        }
    }

    if (fog) {
        ASSERT(ADD_LINE("    "GL_FRAGCOLOR" = mix(color, vec4(fog_color.rgb, color.a), clamp(fog_varying, 0.0, 1.0));\n"));
    } else {
        ASSERT(ADD_LINE("    "GL_FRAGCOLOR" = color;\n"));
    }

    ASSERT(ADD_LINE("}\n"));

    glShaderSource(fragment_shader, num_lines, lines, NULL);

    #undef ADD_LINE
    #undef HIGHP
    #undef LOWP


    return 1;
}

/*-----------------------------------------------------------------------*/

static uint32_t generate_shader_key(
    GLenum primitive_type, int position_count, int texcoord_count,
    GLenum texcolor_type, int tex_offset, int color_count,
    int color_uniform, int fog, int alpha_test, GLenum alpha_comparison)
{
    /* Lists of all valid vertex data counts and texture color types, used
     * to create the shader key.  For data other than position (which is
     * required), a count of 0 indicates that the given field is not
     * present in the vertex data at all; additionally, a texture
     * coordinate count of 0 implies that texturing is disabled. */
    static const uint8_t position_counts[] = {2, 3, 4};
    static const uint8_t texcoord_counts[] = {0, 2};
    static const GLenum texcolor_types[] = {GL_RGBA, GL_ALPHA, 0};
    static const uint8_t color_counts[] = {0, 4};
    static const GLenum alpha_comparison_types[] = {
        GL_LESS, GL_LEQUAL, GL_GEQUAL, GL_GREATER};

    /* Look up the position count, texture coordinate count, texture color
     * type, color count, and alpha comparison type indices. */
    int position_index = -1, texcoord_index = -1, texcolor_index = -1,
        color_index = -1, alpha_comparison_index = -1;
    for (int i = 0; i < lenof(position_counts); i++) {
        if (position_count == position_counts[i]) {
            position_index = i;
            break;
        }
    }
    for (int i = 0; i < lenof(texcoord_counts); i++) {
        if (texcoord_count == texcoord_counts[i]) {
            texcoord_index = i;
            break;
        }
    }
    if (texcoord_count == 0) {
        texcolor_index = 0;  // Unused except to verify parameter validity.
    } else {
        for (int i = 0; i < lenof(texcolor_types); i++) {
            if (texcolor_type == texcolor_types[i]) {
                texcolor_index = i;
                break;
            }
        }
    }
    for (int i = 0; i < lenof(color_counts); i++) {
        if (color_count == color_counts[i]) {
            color_index = i;
            break;
        }
    }
    for (int i = 0; i < lenof(alpha_comparison_types); i++) {
        if (alpha_comparison == alpha_comparison_types[i]) {
            alpha_comparison_index = i;
            break;
        }
    }
    ASSERT(texcolor_index >= 0, return INVALID_SHADER_KEY);
    if (position_index < 0 || texcoord_index < 0 || color_index < 0
     || (alpha_test && alpha_comparison_index < 0)) {
        DLOG("Invalid vertex type: position=%d texcoord=%d texcolor=0x%04X"
             " color=%d alpha_comparison=0x%04X", position_count,
             texcoord_count, texcolor_type, color_count, alpha_comparison);
        return INVALID_SHADER_KEY;
    }

    /* Generate a unique key for the shader by just treating each attribute
     * as a digit in a base-N value (where N varies by digit depending on
     * the number of options for each attribute). */
    uint32_t key = (primitive_type == GL_POINTS ? 1 : 0);
    key *= lenof(position_counts);
    key += position_index;
    key *= 1 + ((lenof(texcoord_counts)-1) * lenof(texcolor_types));
    if (texcoord_index > 0) {
        key += 1 + ((texcoord_index-1)*lenof(texcolor_types) + texcolor_index);
    }
    key *= 2;
    key += tex_offset ? 1 : 0;
    key *= lenof(color_counts);
    key += color_index;
    key *= 2;
    key += color_uniform ? 1 : 0;
    key *= 2;
    key += fog ? 1 : 0;
    key *= 1 + lenof(alpha_comparison_types);
    if (alpha_test) {
        key += 1 + alpha_comparison_index;
    }
    return key;
}

/*-----------------------------------------------------------------------*/

#ifdef DEBUG

static char *get_shader_string(ShaderStringType type, GLuint object)
{
    GLint string_length = -1;
    switch (type) {
      case GET_SHADER_SOURCE:
        glGetShaderiv(object, GL_SHADER_SOURCE_LENGTH, &string_length);
        break;
      case GET_SHADER_INFO:
        glGetShaderiv(object, GL_INFO_LOG_LENGTH, &string_length);
        break;
      case GET_PROGRAM_INFO:
        glGetProgramiv(object, GL_INFO_LOG_LENGTH, &string_length);
        break;
    }
    if (UNLIKELY(string_length < 0)) {
        return mem_strdup("(unavailable)", MEM_ALLOC_TEMP);
    }

    char *buf = mem_alloc(string_length+1, 1, MEM_ALLOC_TEMP);
    if (UNLIKELY(!buf)) {
        /* This will probably fail too, but just in case it works: */
        return mem_strdup("(out of memory)", MEM_ALLOC_TEMP);
    }

    switch (type) {
      case GET_SHADER_SOURCE:
        glGetShaderSource(object, string_length+1, NULL, buf);
        break;
      case GET_SHADER_INFO:
        glGetShaderInfoLog(object, string_length+1, NULL, buf);
        break;
      case GET_PROGRAM_INFO:
        glGetProgramInfoLog(object, string_length+1, NULL, buf);
        break;
    }
    return buf;
}

#endif  // DEBUG

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
