/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/sdl/dyngl.c: OpenGL dynamic loading support for SDL-based
 * environments.
 */

#include "src/base.h"

#define SUPPRESS_GLLOG
/* The type of the string list argument to glShaderSource() varies between
 * "const GLchar **" and const GLchar * const *" depending on the headers
 * in use.  Since there's no way to detect this dynamically and a type
 * conflict is a fatal compiler error, we ignore any provided declaration
 * and add our own (from dyngl-funcs.h, matching the official OpenGL
 * documentation) instead. */
#define glShaderSource glShaderSource_ignored
/* Likewise with the third argument (internalformat) to glTexImage2D(),
 * which varies between "GLint" and "GLenum". */
#define glTexImage2D glTexImage2D_ignored
#include "src/sysdep/opengl/gl-headers.h"
#undef glShaderSource
#undef glTexImage2D
extern void GLAPIENTRY glShaderSource(GLuint, GLsizei, const GLchar **, const GLint *);
extern void GLAPIENTRY glTexImage2D(GLenum, GLint, GLint, GLsizei, GLsizei, GLint, GLenum, GLenum, const GLvoid *);

#include "src/sysdep/sdl/dyngl.h"

#include <SDL.h>

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Have we already loaded the OpenGL library? */
static uint8_t loaded;

/* Which extensions are available? */
static uint8_t has_compressed_textures;
static uint8_t has_vertex_buffers;
static uint8_t has_shaders;
static uint8_t has_framebuffers;
static uint8_t has_opengl3;

/*-----------------------------------------------------------------------*/

/* Function pointers for all GL functions. */

#ifndef GLAPIENTRY
# define GLAPIENTRY /*nothing*/
#endif

#define DYNGL_FUNC(rettype,name,altname,params,args,required,category) \
    static rettype (GLAPIENTRY *p##name) params;
#include "src/sysdep/opengl/dyngl-funcs.h"
#undef DYNGL_FUNC

/*-----------------------------------------------------------------------*/

#ifdef INCLUDE_TESTS

/* Override for glGetString() set from TEST_dyngl_override_glGetString(). */
static const GLubyte * GLAPIENTRY (*glGetString_override)(
    GLenum, const GLubyte * GLAPIENTRY (*)(GLenum)) = NULL;

/* Original glGetString() function pointer when overridden. */
static const GLubyte * GLAPIENTRY (*original_glGetString)(GLenum);

/* Wrapper for the override function which passes the original function
 * pointer. */
static const GLubyte * GLAPIENTRY wrap_glGetString_override(GLenum name) {
    return (*glGetString_override)(name, original_glGetString);
}

#endif

/*************************************************************************/
/************** Interface routine (called from graphics.c) ***************/
/*************************************************************************/

int dyngl_init(void)
{
    if (loaded) {
        return 1;
    }

    pglGetString = SDL_GL_GetProcAddress("glGetString");
    if (UNLIKELY(!pglGetString)) {
        DLOG("glGetString() not found!");
        return 0;
    }
    const char *gl_version = (const char *)glGetString(GL_VERSION);
    if (UNLIKELY(!gl_version || !*gl_version)) {
        DLOG("Failed to get GL version!");
        return 0;
    }
    const char *gl_subversion;
    int major = (int)strtoul(gl_version, (char **)&gl_subversion, 10);
    if (UNLIKELY(*gl_subversion != '.')) {
        DLOG("Failed to parse OpenGL version string!");
        return 0;
    }
    int minor = (int)strtoul(gl_subversion+1, NULL, 10);

    has_compressed_textures = 1;
    has_vertex_buffers = 1;
    has_shaders = 1;
    has_framebuffers = 1;
    has_opengl3 = 1;

    UNUSED int has_;  // Avoid errors from an empty category on base functions.
#define DYNGL_FUNC(rettype,name,altname,params,args,required,category)  \
    p##name = SDL_GL_GetProcAddress(#name);                             \
    if (strcmp(#altname, #name) != 0 && !p##name) {                     \
        p##name = SDL_GL_GetProcAddress(#altname);                      \
    }                                                                   \
    if (!p##name) {                                                     \
        if (required) {                                                 \
            DLOG("Failed to get address of %s: %s", #name, SDL_GetError()); \
            return 0;                                                   \
        } else {                                                        \
            has_##category = 0;                                         \
        }                                                               \
    }
#include "src/sysdep/opengl/dyngl-funcs.h"
#undef DYNGL_FUNC

#ifdef INCLUDE_TESTS
    if (glGetString_override) {
        original_glGetString = pglGetString;
        pglGetString = wrap_glGetString_override;
    }
#endif

    loaded = 1;
    return 1;
}

/*-----------------------------------------------------------------------*/

void dyngl_cleanup(void)
{
#define DYNGL_FUNC(rettype,name,altname,params,args,required,category) \
    p##name = NULL;
#include "src/sysdep/opengl/dyngl-funcs.h"
#undef DYNGL_FUNC

#ifdef INCLUDE_TESTS
    glGetString_override = NULL;
#endif

    loaded = 0;
}

/*-----------------------------------------------------------------------*/

int dyngl_has_compressed_textures(void)
{
    return has_compressed_textures;
}

/*-----------------------------------------------------------------------*/

int dyngl_has_vertex_buffers(void)
{
    return has_vertex_buffers;
}

/*-----------------------------------------------------------------------*/

int dyngl_has_shaders(void)
{
    return has_shaders;
}

/*-----------------------------------------------------------------------*/

int dyngl_has_framebuffers(void)
{
    return has_framebuffers;
}

/*-----------------------------------------------------------------------*/

int dyngl_has_opengl3(void)
{
    return has_opengl3;
}

/*************************************************************************/
/************************* Test control routines *************************/
/*************************************************************************/

#ifdef INCLUDE_TESTS

extern void TEST_dyngl_override_glGetString(
    const GLubyte * GLAPIENTRY (*function)(
        GLenum name,
        const GLubyte * GLAPIENTRY (*original_glGetString)(GLenum)))
{
    glGetString_override = function;
}

#endif  // INCLUDE_TESTS

/*************************************************************************/
/************************* OpenGL stub functions *************************/
/*************************************************************************/

#define DYNGL_FUNC(rettype,name,altname,params,args,required,category) \
    rettype GLAPIENTRY name params {return (*p##name) args;}
#include "src/sysdep/opengl/dyngl-funcs.h"
#undef DYNGL_FUNC

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
