/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/graphics/internal.c: Helper functions for the graphics
 * subsystem tests.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/graphics.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"

/*************************************************************************/
/*************************************************************************/

int run_tests_in_window(int (*function)(void))
{
    return run_tests_in_sized_window(function, TESTW, TESTH);
}

/*-----------------------------------------------------------------------*/

int run_tests_in_sized_window(int (*function)(void), int width, int height)
{
    ASSERT(graphics_init());
    ASSERT(graphics_set_display_attr("stencil_bits", 8));  // Ensure a stencil buffer.
    graphics_set_display_attr("vsync", 0);  // Run as fast as possible.
    if (!open_window(width, height)) {
        graphics_cleanup();
        FAIL("Unable to open window for tests");
    }
    graphics_set_viewport(0, 0, width, height);

    const int result = (*function)();

#ifdef SIL_PLATFORM_ANDROID
    /* Some devices/versions crash if we don't have a single rendering
     * operation between eglCreateContext() and eglDestroyContext(). */
    graphics_start_frame();
    graphics_clear(0, 0, 0);
    graphics_finish_frame();
#endif
    graphics_cleanup();
    return result;
}

/*-----------------------------------------------------------------------*/

int open_window(int width, int height)
{
    if (!graphics_set_display_attr("window", 1)) {
        /* Look for the smallest usable display size and use that instead. */
        const GraphicsDisplaySizeList *size_list;
        ASSERT(size_list = graphics_list_display_sizes());
        ASSERT(size_list->num_sizes > 0);
        int i;
        for (i = 0; i < size_list->num_sizes; i++) {
            if (size_list->sizes[i].width >= width
             && size_list->sizes[i].height >= height) {
                break;
            }
        }
        if (i >= size_list->num_sizes) {
            FAIL("No available display mode of size at least %dx%d",
                 width, height);
        }
        width = size_list->sizes[i].width;
        height = size_list->sizes[i].height;
    }
    return graphics_set_display_mode(width, height);
}

/*-----------------------------------------------------------------------*/

uint8_t *grab_display(void)
{
    const int width = TESTW, height = TESTH;

    uint8_t *pixels = mem_alloc(width*height*4, 0, MEM_ALLOC_TEMP);
    if (!pixels) {
        DLOG("Failed to allocate memory");
        return NULL;
    }
    if (!graphics_read_pixels(0, 0, width, height, pixels)) {
        DLOG("Failed to read pixels");
        mem_free(pixels);
        return NULL;
    }
    return pixels;
}

/*-----------------------------------------------------------------------*/

void draw_square(float z, float r, float g, float b, float a)
{
    ASSERT(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    ASSERT(graphics_add_vertex(&(Vector3f){-0.5, -0.5, z}, NULL,
                               &(Vector4f){r, g, b, a}));
    ASSERT(graphics_add_vertex(&(Vector3f){-0.5, +0.5, z}, NULL,
                               &(Vector4f){r, g, b, a}));
    ASSERT(graphics_add_vertex(&(Vector3f){+0.5, +0.5, z}, NULL,
                               &(Vector4f){r, g, b, a}));
    ASSERT(graphics_add_vertex(&(Vector3f){+0.5, -0.5, z}, NULL,
                               &(Vector4f){r, g, b, a}));
    ASSERT(graphics_end_and_draw_primitive());
}

/*-----------------------------------------------------------------------*/

int check_colored_rectangle(int w, int h, int cx, int cy,
                            float r, float g, float b)
{
    const int R = iroundf(r*255), G = iroundf(g*255), B = iroundf(b*255);

    uint8_t pixels[64*64*4];
    ASSERT(graphics_read_pixels(0, 0, 64, 64, pixels));
    for (int i = 0; i < 64*64*4; i += 4) {
        const int x = (i/4) % 64;
        const int y = (i/4) / 64;
        if (x >= cx-w/2 && x < cx+w/2 && y >= cy-h/2 && y < cy+h/2) {
            CHECK_PIXEL_NEAR(&pixels[i], R,G,B,255, 1, x, y);
        } else {
            CHECK_PIXEL(&pixels[i], 0,0,0,255, x, y);
        }
    }
    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
