/*
 * System Interface Library for games
 * Copyright (c) 2011 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/psp/userdata.c: Tests for PSP implementation of the user
 * data access functions.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/graphics.h"
#include "src/memory.h"
#include "src/resource.h"
#include "src/resource/package.h"
#include "src/sysdep.h"
#include "src/sysdep/psp/internal.h"
#include "src/sysdep/test.h"
#include "src/test/base.h"
#include "src/texture.h"
#include "src/thread.h"
#include "src/userdata.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Parameters for the save file we operate on. */
#define SAVE_PROGRAM_NAME  "TEST00000SIL"
#define SAVE_NUM           1

/* Pathnames corresponding to the parameters above. */
#define PATH_SAVE_DIR       "ms0:/PSP/SAVEDATA/" SAVE_PROGRAM_NAME "_001"
#define PATH_SAVE_FILE      PATH_SAVE_DIR "/save.bin"
#define PATH_SETTINGS_DIR   "ms0:/PSP/SAVEDATA/" SAVE_PROGRAM_NAME "_Settings"
#define PATH_SETTINGS_FILE  PATH_SETTINGS_DIR "/settings.bin"
#define PATH_STATS_DIR      "ms0:/PSP/SAVEDATA/" SAVE_PROGRAM_NAME "_Stats"
#define PATH_STATS_FILE     PATH_STATS_DIR "/stats.bin"

/*************************************************************************/
/********************** Helper macros and functions **********************/
/*************************************************************************/

/**
 * CHECK_USERDATA_MEMORY_FAILURES:  Check that the given userdata
 * operation handles memory allocation failures properly.  Similar to the
 * CHECK_MEMORY_FAILURES() macro, except that this macro waits for the
 * operation to complete.
 *
 * Call like:
 *    CHECK_USERDATA_MEMORY_FAILURES(userdata_save_savefile(...))
 * (the parameter should return an operation ID).
 */
#define CHECK_USERDATA_MEMORY_FAILURES(operation) \
    CHECK_MEMORY_FAILURES( \
        (id = (operation)) && (userdata_wait(id), userdata_get_result(id)))

/*-----------------------------------------------------------------------*/

/**
 * path_exists:  Return whether the given path names an existing file or
 * directory.
 *
 * [Parameters]
 *     path: Pathname to check.
 * [Return value]
 *     True if the path names an existing file or directory, false if not.
 */
static int path_exists(const char *path)
{
    SceIoStat st;
    return sceIoGetstat(path, &st) == 0;
}

/*-----------------------------------------------------------------------*/

/**
 * remove_or_warn:  Attempt to remove the given file or directory, logging
 * a warning if the file or directory exists but cannot be removed.
 *
 * [Parameters]
 *     path: Pathname of file or directory to remove.
 *     is_dir: True if the item is a directory, false if a file.
 */
#define remove_or_warn(path,is_dir) \
    remove_or_warn_((path), (is_dir), __FILE__, __LINE__, __FUNCTION__)
static void remove_or_warn_(const char *path, int is_dir,
                            const char *file, int line, const char *function)
{
    uint32_t error;
    if (is_dir) {
        error = sceIoRmdir(path);
    } else {
        error = sceIoRemove(path);
    }
    if (error != 0 && error != PSP_ENOENT) {
        testlog_log(file, line, function, TESTLOG_WARN,
                    "WARN: Failed to remove%s %s: %s",
                    is_dir ? " directory" : "", path, psp_strerror(error));
    }
}

/*************************************************************************/
/******************* ICON0.PNG resource package module *******************/
/*************************************************************************/

/*
 * This module is used to implement a virtual resource at ICON0.PNG (the
 * actual file is located at testdata/psp/ICON0.PNG).
 */

/* Internal data for the package module. */
typedef struct IconPackagePrivate IconPackagePrivate;
struct IconPackagePrivate {
    SysFile *fh;  // Persistent file handle for icon file.
};

/*-----------------------------------------------------------------------*/

static int icon0_init(PackageModuleInfo *module)
{
    IconPackagePrivate *private;
    ASSERT(private = mem_alloc(sizeof(*private), 0, 0));
    ASSERT(private->fh = sys_file_open("testdata/psp/ICON0.PNG"));
    module->module_data = private;
    return 1;
}

/*-----------------------------------------------------------------------*/

static void icon0_cleanup(PackageModuleInfo *module)
{
    IconPackagePrivate *private = (IconPackagePrivate *)module->module_data;
    sys_file_close(private->fh);
    mem_free(private);
}

/*-----------------------------------------------------------------------*/

static int icon0_file_info(PackageModuleInfo *module,
                             const char *path, SysFile **file_ret,
                             int64_t *pos_ret, int *len_ret,
                             int *comp_ret, int *size_ret)
{
    IconPackagePrivate *private = (IconPackagePrivate *)module->module_data;

    if (strcmp(path, "ICON0.PNG") != 0) {
        return 0;
    }

    *file_ret = private->fh;
    *pos_ret = 0;
    *len_ret = sys_file_size(private->fh);
    *comp_ret = 0;
    *size_ret = *len_ret;
    return 1;
}

/*-----------------------------------------------------------------------*/

static int icon0_decompress(
    UNUSED PackageModuleInfo *module, UNUSED void *state,
    UNUSED const void *in, UNUSED int insize,
    UNUSED void *out, UNUSED int outsize)
{
    return 0;
}

/*-----------------------------------------------------------------------*/

static PackageModuleInfo icon0_module = {
    .prefix     = "",
    .init       = icon0_init,
    .cleanup    = icon0_cleanup,
    .file_info  = icon0_file_info,
    .decompress = icon0_decompress,
};

/*************************************************************************/
/***************** Test runner and init/cleanup routines *****************/
/*************************************************************************/

static int do_test_psp_userdata(void);
int test_psp_userdata(void)
{
    /* Make sure the test files we're about to create don't already exist. */
    if (path_exists(PATH_SAVE_DIR)
     || path_exists(PATH_SETTINGS_DIR)
     || path_exists(PATH_STATS_DIR)) {
        FAIL("Test would overwrite existing save files!  Ensure that the"
             " following directories do not exist:"
             "\n    " PATH_SAVE_DIR
             "\n    " PATH_SETTINGS_DIR
             "\n    " PATH_STATS_DIR);
    }

    /* Run the tests. */
    DLOG("Running PSP userdata tests (slow)...");
    CHECK_TRUE(graphics_init());
    resource_init();
    sys_test_userdata_use_live_routines = 1;
    const int result = do_test_psp_userdata();
    sys_test_userdata_use_live_routines = 0;
    resource_cleanup();
    graphics_cleanup();

    /* Remove any leftover save files before returning. */
    remove_or_warn(PATH_SAVE_FILE, 0);
    remove_or_warn(PATH_SAVE_DIR "/PARAM.SFO", 0);
    remove_or_warn(PATH_SAVE_DIR "/ICON0.PNG", 0);
    remove_or_warn(PATH_SAVE_DIR, 1);
    remove_or_warn(PATH_SETTINGS_FILE, 0);
    remove_or_warn(PATH_SETTINGS_DIR "/PARAM.SFO", 0);
    remove_or_warn(PATH_SETTINGS_DIR "/ICON0.PNG", 0);
    remove_or_warn(PATH_SETTINGS_DIR, 1);
    remove_or_warn(PATH_STATS_FILE, 0);
    remove_or_warn(PATH_STATS_DIR "/PARAM.SFO", 0);
    remove_or_warn(PATH_STATS_DIR "/ICON0.PNG", 0);
    remove_or_warn(PATH_STATS_DIR, 1);

    return result;
}

DEFINE_GENERIC_TEST_RUNNER(do_test_psp_userdata)

/*-----------------------------------------------------------------------*/

TEST_INIT(init)
{
    TEST_userdata_reset();
    userdata_set_program_name(SAVE_PROGRAM_NAME, "Userdata Test");
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST_CLEANUP(cleanup)
{
    TEST_userdata_reset();

    /* Free any memory allocated to statistics to avoid false-positive
     * leak reports. */
    static const struct UserStatInfo stats[] = {
        {.id = 0, .type = USERDATA_STAT_FLAG}};
    TEST_mem_fail_after(6, 1);  // Carefully tuned to fail at the right place.
    int id;
    CHECK_TRUE(id = userdata_register_stats(stats, lenof(stats)));
    userdata_wait(id);
    CHECK_FALSE(userdata_get_result(id));
    TEST_mem_fail_after(-1, 0);

    return 1;
}

/*************************************************************************/
/**************************** Save data tests ****************************/
/*************************************************************************/

TEST(test_savefile)
{
    int id;
    void *data;
    uint32_t size;
    int image;

    CHECK_TRUE(id = userdata_save_savefile(
                   SAVE_NUM, "111", 3, "title", "desc", 0));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));

    CHECK_TRUE(id = userdata_load_savefile(SAVE_NUM, &data, &size, &image));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_INTEQUAL(size, 3);
    CHECK_MEMEQUAL(data, "111", 3);
    CHECK_FALSE(image);
    mem_free(data);

    CHECK_TRUE(path_exists(PATH_SAVE_DIR));
    CHECK_TRUE(path_exists(PATH_SAVE_FILE));
    CHECK_TRUE(path_exists(PATH_SAVE_DIR "/PARAM.SFO"));
    /* We should not have an ICON0.PNG because we didn't pass in an image
     * and we don't have a default ICON0.PNG resource. */
    CHECK_FALSE(path_exists(PATH_SAVE_DIR "/ICON0.PNG"));

    remove_or_warn(PATH_SAVE_FILE, 0);
    remove_or_warn(PATH_SAVE_DIR "/PARAM.SFO", 0);
    remove_or_warn(PATH_SAVE_DIR, 1);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_savefile_image)
{
    int texture;
    ASSERT(texture = texture_create(144, 80, 0, 0));
    uint8_t *pixels;
    ASSERT(pixels = texture_lock(texture));
    for (int i = 0; i < 144*80; i++) {
        const int y = i/144;
        pixels[i*4+0] = 1 * (y+1);
        pixels[i*4+1] = 2 * (y+1);
        pixels[i*4+2] = 3 * (y+1);
        pixels[i*4+3] = 0xFF;
    }
    texture_unlock(texture);

    int id;
    void *data;
    uint32_t size;
    int image;

    CHECK_TRUE(id = userdata_save_savefile(
                   SAVE_NUM, "111", 3, "title", "desc", texture));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    texture_destroy(texture);

    CHECK_TRUE(id = userdata_load_savefile(SAVE_NUM, &data, &size, &image));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_INTEQUAL(size, 3);
    CHECK_MEMEQUAL(data, "111", 3);
    CHECK_TRUE(image);
    CHECK_INTEQUAL(texture_width(image), 144);
    CHECK_INTEQUAL(texture_height(image), 80);
    CHECK_FLOATEQUAL(texture_scale(image), 1.0f);
    const uint8_t *image_pixels;
    CHECK_TRUE(image_pixels = (uint8_t *)texture_lock_readonly(image));
    for (int i = 0; i < 144*80; i++) {
        const int y = i/144;
        CHECK_INTEQUAL(image_pixels[i*4+0], 1 * (y+1));
        CHECK_INTEQUAL(image_pixels[i*4+1], 2 * (y+1));
        CHECK_INTEQUAL(image_pixels[i*4+2], 3 * (y+1));
        CHECK_INTEQUAL(image_pixels[i*4+3], 0xFF);
    }
    mem_free(data);
    texture_destroy(image);

    CHECK_TRUE(path_exists(PATH_SAVE_DIR));
    CHECK_TRUE(path_exists(PATH_SAVE_FILE));
    CHECK_TRUE(path_exists(PATH_SAVE_DIR "/PARAM.SFO"));
    CHECK_TRUE(path_exists(PATH_SAVE_DIR "/ICON0.PNG"));

    remove_or_warn(PATH_SAVE_FILE, 0);
    remove_or_warn(PATH_SAVE_DIR "/PARAM.SFO", 0);
    remove_or_warn(PATH_SAVE_DIR "/ICON0.PNG", 0);
    remove_or_warn(PATH_SAVE_DIR, 1);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_savefile_default_icon)
{
    int id;
    void *data;
    uint32_t size;
    int image;

    CHECK_TRUE(resource_register_package(&icon0_module));
    CHECK_TRUE(id = userdata_save_savefile(
                   SAVE_NUM, "111", 3, "title", "desc", 0));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    resource_unregister_package(&icon0_module);

    CHECK_TRUE(id = userdata_load_savefile(SAVE_NUM, &data, &size, &image));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_INTEQUAL(size, 3);
    CHECK_MEMEQUAL(data, "111", 3);
    /* We should get no image back even though we have a default icon
     * because the PSP userdata code only parses PNGs that it generated. */
    CHECK_FALSE(image);
    mem_free(data);

    CHECK_TRUE(path_exists(PATH_SAVE_DIR));
    CHECK_TRUE(path_exists(PATH_SAVE_FILE));
    CHECK_TRUE(path_exists(PATH_SAVE_DIR "/PARAM.SFO"));
    /* We should now get an ICON0.PNG identical to our test icon. */
    CHECK_TRUE(path_exists(PATH_SAVE_DIR "/ICON0.PNG"));

    void *expected_data;
    int expected_size;
    ResourceManager *resmgr = resource_create(2);
    int resid;
    ASSERT(resid = resource_load_data(resmgr, "testdata/psp/ICON0.PNG", 0, 0));
    resource_wait(resmgr, resource_mark(resmgr));
    ASSERT(data = resource_get_data(resmgr, resid, &expected_size));
    ASSERT(expected_data = mem_alloc(expected_size, 0, 0));
    memcpy(expected_data, data, expected_size);
    resource_destroy(resmgr);

    void *actual_data;
    int actual_size;
    SysFile *fh;
    ASSERT(fh = sys_file_open(PATH_SAVE_DIR "/ICON0.PNG"));
    actual_size = sys_file_size(fh);
    ASSERT(actual_data = mem_alloc(lbound(actual_size,1), 0, 0));
    ASSERT(sys_file_read(fh, actual_data, actual_size) == actual_size);
    sys_file_close(fh);

    CHECK_INTEQUAL(actual_size, expected_size);
    CHECK_MEMEQUAL(actual_data, expected_data, expected_size);
    mem_free(actual_data);
    mem_free(expected_data);

    remove_or_warn(PATH_SAVE_FILE, 0);
    remove_or_warn(PATH_SAVE_DIR "/PARAM.SFO", 0);
    remove_or_warn(PATH_SAVE_DIR "/ICON0.PNG", 0);
    remove_or_warn(PATH_SAVE_DIR, 1);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_savefile_image_memory_failure)
{
    int texture;
    ASSERT(texture = texture_create(144, 80, 0, 0));
    uint8_t *pixels;
    ASSERT(pixels = texture_lock(texture));
    for (int i = 0; i < 144*80; i++) {
        const int y = i/144;
        pixels[i*4+0] = 1 * (y+1);
        pixels[i*4+1] = 2 * (y+1);
        pixels[i*4+2] = 3 * (y+1);
        pixels[i*4+3] = 0xFF;
    }
    texture_unlock(texture);

    int id;
    void *data;
    uint32_t size;
    int image;

    CHECK_USERDATA_MEMORY_FAILURES(
        userdata_save_savefile(SAVE_NUM, "111", 3, "title", "desc", texture));
    texture_destroy(texture);

    CHECK_USERDATA_MEMORY_FAILURES(
        userdata_load_savefile(SAVE_NUM, &data, &size, &image));
    CHECK_INTEQUAL(size, 3);
    CHECK_MEMEQUAL(data, "111", 3);
    CHECK_FALSE(image);  // Should be missing due to memory allocation failure.
    mem_free(data);

    CHECK_TRUE(path_exists(PATH_SAVE_DIR));
    CHECK_TRUE(path_exists(PATH_SAVE_FILE));
    CHECK_TRUE(path_exists(PATH_SAVE_DIR "/PARAM.SFO"));
    CHECK_FALSE(path_exists(PATH_SAVE_DIR "/ICON0.PNG"));

    remove_or_warn(PATH_SAVE_FILE, 0);
    remove_or_warn(PATH_SAVE_DIR "/PARAM.SFO", 0);
    remove_or_warn(PATH_SAVE_DIR "/ICON0.PNG", 0);
    remove_or_warn(PATH_SAVE_DIR, 1);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_savefile_default_icon_memory_failure)
{
    int id;
    void *data;
    uint32_t size;
    int image;

    CHECK_TRUE(resource_register_package(&icon0_module));
    CHECK_USERDATA_MEMORY_FAILURES(
        userdata_save_savefile(SAVE_NUM, "111", 3, "title", "desc", 0));
    resource_unregister_package(&icon0_module);

    CHECK_USERDATA_MEMORY_FAILURES(
        userdata_load_savefile(SAVE_NUM, &data, &size, &image));
    CHECK_INTEQUAL(size, 3);
    CHECK_MEMEQUAL(data, "111", 3);
    CHECK_FALSE(image);
    mem_free(data);

    CHECK_TRUE(path_exists(PATH_SAVE_DIR));
    CHECK_TRUE(path_exists(PATH_SAVE_FILE));
    CHECK_TRUE(path_exists(PATH_SAVE_DIR "/PARAM.SFO"));
    CHECK_FALSE(path_exists(PATH_SAVE_DIR "/ICON0.PNG"));

    remove_or_warn(PATH_SAVE_FILE, 0);
    remove_or_warn(PATH_SAVE_DIR "/PARAM.SFO", 0);
    remove_or_warn(PATH_SAVE_DIR "/ICON0.PNG", 0);
    remove_or_warn(PATH_SAVE_DIR, 1);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_savefile_missing)
{
    int id;
    void *data;
    uint32_t size;

    CHECK_TRUE(id = userdata_load_savefile(SAVE_NUM+1, &data, &size, NULL));
    userdata_wait(id);
    CHECK_FALSE(userdata_get_result(id));

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_settings)
{
    int id;
    void *data;
    uint32_t size;

    CHECK_TRUE(id = userdata_save_settings("2222", 4, "title", "desc"));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));

    CHECK_TRUE(id = userdata_load_settings(&data, &size));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_INTEQUAL(size, 4);
    CHECK_MEMEQUAL(data, "2222", 4);
    mem_free(data);

    CHECK_TRUE(path_exists(PATH_SETTINGS_DIR));
    CHECK_TRUE(path_exists(PATH_SETTINGS_FILE));
    CHECK_TRUE(path_exists(PATH_SETTINGS_DIR "/PARAM.SFO"));
    CHECK_FALSE(path_exists(PATH_SETTINGS_DIR "/ICON0.PNG"));

    remove_or_warn(PATH_SETTINGS_FILE, 0);
    remove_or_warn(PATH_SETTINGS_DIR "/PARAM.SFO", 0);
    remove_or_warn(PATH_SETTINGS_DIR, 1);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Identical to the same-named test in src/test/userdata.c, with the
 * addition of presence tests for and removal of Memory Stick files. */
TEST(test_stats)
{
    static const struct UserStatInfo stats[] = {
        {.id = 0, .type = USERDATA_STAT_FLAG},
        {.id = 2, .type = USERDATA_STAT_UINT32},
        {.id = 3, .type = USERDATA_STAT_DOUBLE},
        {.id = 5, .type = USERDATA_STAT_UINT32_MAX},
        {.id = 7, .type = USERDATA_STAT_DOUBLE_MAX},
    };
    struct UserStatValue values_in[] = {
        {.id = 0},
        {.id = 2},
        {.id = 3},
        {.id = 5},
        {.id = 7},
    };
    struct UserStatValue values_out[] = {
        {.id = 2},
        {.id = 5},
        {.id = 3},
        {.id = 7},
        {.id = 0},
    };
    int id;

    /* Register stats, and make sure they're all initialized to zero. */
    CHECK_TRUE(id = userdata_register_stats(stats, lenof(stats)));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_TRUE(id = userdata_retrieve_stats(values_out, lenof(values_out)));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_DOUBLEEQUAL(values_out[0].value, 0);
    CHECK_DOUBLEEQUAL(values_out[1].value, 0);
    CHECK_DOUBLEEQUAL(values_out[2].value, 0);
    CHECK_DOUBLEEQUAL(values_out[3].value, 0);
    CHECK_DOUBLEEQUAL(values_out[4].value, 0);

    /* Set values, and make sure they're properly set/retrieved by ID
     * rather than array index. */
    values_in[0].value = -999;  // Flag, should turn into 1.
    values_in[1].value = 100;
    values_in[2].value = 1e10;
    values_in[3].value = 100;
    values_in[4].value = 1e10;
    CHECK_TRUE(id = userdata_update_stats(values_in, lenof(values_in)));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_TRUE(id = userdata_retrieve_stats(values_out, lenof(values_out)));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_DOUBLEEQUAL(values_out[0].value, 100);
    CHECK_DOUBLEEQUAL(values_out[1].value, 100);
    CHECK_DOUBLEEQUAL(values_out[2].value, 1e10);
    CHECK_DOUBLEEQUAL(values_out[3].value, 1e10);
    CHECK_DOUBLEEQUAL(values_out[4].value, 1);

    /* Make sure attempts to reset flags or lower _MAX values fail. */
    values_in[0].value = 0;
    values_in[1].value = 10;
    values_in[2].value = 1e5;
    values_in[3].value = 10;
    values_in[4].value = 1e5;
    CHECK_TRUE(id = userdata_update_stats(values_in, lenof(values_in)));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_TRUE(id = userdata_retrieve_stats(values_out, lenof(values_out)));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_DOUBLEEQUAL(values_out[0].value, 10);
    CHECK_DOUBLEEQUAL(values_out[1].value, 100);
    CHECK_DOUBLEEQUAL(values_out[2].value, 1e5);
    CHECK_DOUBLEEQUAL(values_out[3].value, 1e10);
    CHECK_DOUBLEEQUAL(values_out[4].value, 1);

    /* Make sure setting a flag that is already set does nothing. */
    values_in[0].value = 1;
    CHECK_TRUE(id = userdata_update_stats(values_in, 1));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_TRUE(id = userdata_retrieve_stats(values_out, lenof(values_out)));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_DOUBLEEQUAL(values_out[4].value, 1);

    /* Check double->int conversion for UINT32 values. */
    values_in[1].value = 123.4;
    values_in[3].value = 1e10;
    CHECK_TRUE(id = userdata_update_stats(&values_in[1], 3));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_TRUE(id = userdata_retrieve_stats(values_out, lenof(values_out)));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_DOUBLEEQUAL(values_out[0].value, 123);
    CHECK_DOUBLEEQUAL(values_out[1].value, 0xFFFFFFFFU);
    CHECK_DOUBLEEQUAL(values_out[2].value, 1e5);
    CHECK_DOUBLEEQUAL(values_out[3].value, 1e10);
    CHECK_DOUBLEEQUAL(values_out[4].value, 1);

    values_in[1].value = -123.4;
    CHECK_TRUE(id = userdata_update_stats(&values_in[1], 1));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_TRUE(id = userdata_retrieve_stats(&values_out[0], 1));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_DOUBLEEQUAL(values_out[0].value, 0);

    /* Check that clear() clears everything, including flags and _MAX stats. */
    CHECK_TRUE(id = userdata_clear_stats());
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_TRUE(id = userdata_retrieve_stats(values_out, lenof(values_out)));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_DOUBLEEQUAL(values_out[0].value, 0);
    CHECK_DOUBLEEQUAL(values_out[1].value, 0);
    CHECK_DOUBLEEQUAL(values_out[2].value, 0);
    CHECK_DOUBLEEQUAL(values_out[3].value, 0);
    CHECK_DOUBLEEQUAL(values_out[4].value, 0);

    /* Check that a second clear() does nothing (and doesn't break). */
    CHECK_TRUE(id = userdata_clear_stats());
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_TRUE(id = userdata_retrieve_stats(values_out, lenof(values_out)));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_DOUBLEEQUAL(values_out[0].value, 0);
    CHECK_DOUBLEEQUAL(values_out[1].value, 0);
    CHECK_DOUBLEEQUAL(values_out[2].value, 0);
    CHECK_DOUBLEEQUAL(values_out[3].value, 0);
    CHECK_DOUBLEEQUAL(values_out[4].value, 0);

    /* Check that a second register succeeds and preserves values. */
    values_in[0].value = 1;
    values_in[1].value = 100;
    values_in[2].value = 1e10;
    values_in[3].value = 100;
    values_in[4].value = 1e10;
    CHECK_TRUE(id = userdata_update_stats(values_in, lenof(values_in)));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_TRUE(id = userdata_retrieve_stats(values_out, lenof(values_out)));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_DOUBLEEQUAL(values_out[0].value, 100);
    CHECK_DOUBLEEQUAL(values_out[1].value, 100);
    CHECK_DOUBLEEQUAL(values_out[2].value, 1e10);
    CHECK_DOUBLEEQUAL(values_out[3].value, 1e10);
    CHECK_DOUBLEEQUAL(values_out[4].value, 1);
    CHECK_TRUE(id = userdata_register_stats(stats, lenof(stats)));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_TRUE(id = userdata_retrieve_stats(values_out, lenof(values_out)));
    userdata_wait(id);
    CHECK_TRUE(userdata_get_result(id));
    CHECK_DOUBLEEQUAL(values_out[0].value, 100);
    CHECK_DOUBLEEQUAL(values_out[1].value, 100);
    CHECK_DOUBLEEQUAL(values_out[2].value, 1e10);
    CHECK_DOUBLEEQUAL(values_out[3].value, 1e10);
    CHECK_DOUBLEEQUAL(values_out[4].value, 1);

    CHECK_TRUE(path_exists(PATH_STATS_DIR));
    CHECK_TRUE(path_exists(PATH_STATS_FILE));
    CHECK_TRUE(path_exists(PATH_STATS_DIR "/PARAM.SFO"));
    CHECK_FALSE(path_exists(PATH_STATS_DIR "/ICON0.PNG"));

    remove_or_warn(PATH_STATS_FILE, 0);
    remove_or_warn(PATH_STATS_DIR "/PARAM.SFO", 0);
    remove_or_warn(PATH_STATS_DIR, 1);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Identical to the same-named test in src/test/userdata.c, with the
 * addition of presence tests for and removal of Memory Stick files. */
TEST(test_stats_memory_failure)
{
    /* Same as (portions of) test_stats(), except that we bracket all
     * userdata calls with CHECK_USERDATA_MEMORY_FAILURES(). */

    static const struct UserStatInfo stats[] = {
        {.id = 0, .type = USERDATA_STAT_FLAG},
        {.id = 2, .type = USERDATA_STAT_UINT32},
        {.id = 3, .type = USERDATA_STAT_DOUBLE},
        {.id = 5, .type = USERDATA_STAT_UINT32_MAX},
        {.id = 7, .type = USERDATA_STAT_DOUBLE_MAX},
    };
    struct UserStatValue values_in[] = {
        {.id = 0},
        {.id = 2},
        {.id = 3},
        {.id = 5},
        {.id = 7},
    };
    struct UserStatValue values_out[] = {
        {.id = 2},
        {.id = 5},
        {.id = 3},
        {.id = 7},
        {.id = 0},
    };
    int id;

    /* Register stats, and make sure they're all initialized to zero. */
    CHECK_USERDATA_MEMORY_FAILURES(
        userdata_register_stats(stats, lenof(stats)));
    CHECK_USERDATA_MEMORY_FAILURES(
        userdata_retrieve_stats(values_out, lenof(values_out)));
    CHECK_DOUBLEEQUAL(values_out[0].value, 0);
    CHECK_DOUBLEEQUAL(values_out[1].value, 0);
    CHECK_DOUBLEEQUAL(values_out[2].value, 0);
    CHECK_DOUBLEEQUAL(values_out[3].value, 0);
    CHECK_DOUBLEEQUAL(values_out[4].value, 0);

    /* Set values, and make sure they're properly set/retrieved by ID
     * rather than array index. */
    values_in[0].value = -999;  // Flag, should turn into 1.
    values_in[1].value = 100;
    values_in[2].value = 1e10;
    values_in[3].value = 100;
    values_in[4].value = 1e10;
    CHECK_USERDATA_MEMORY_FAILURES(
        userdata_update_stats(values_in, lenof(values_in)));
    CHECK_USERDATA_MEMORY_FAILURES(
        userdata_retrieve_stats(values_out, lenof(values_out)));
    CHECK_DOUBLEEQUAL(values_out[0].value, 100);
    CHECK_DOUBLEEQUAL(values_out[1].value, 100);
    CHECK_DOUBLEEQUAL(values_out[2].value, 1e10);
    CHECK_DOUBLEEQUAL(values_out[3].value, 1e10);
    CHECK_DOUBLEEQUAL(values_out[4].value, 1);

    /* Check that clear() clears everything, including flags and _MAX stats. */
    CHECK_USERDATA_MEMORY_FAILURES(userdata_clear_stats());
    CHECK_USERDATA_MEMORY_FAILURES(
        userdata_retrieve_stats(values_out, lenof(values_out)));
    CHECK_DOUBLEEQUAL(values_out[0].value, 0);
    CHECK_DOUBLEEQUAL(values_out[1].value, 0);
    CHECK_DOUBLEEQUAL(values_out[2].value, 0);
    CHECK_DOUBLEEQUAL(values_out[3].value, 0);
    CHECK_DOUBLEEQUAL(values_out[4].value, 0);

    /* Check that a second register succeeds and preserves values. */
    values_in[0].value = 1;
    values_in[1].value = 100;
    values_in[2].value = 1e10;
    values_in[3].value = 100;
    values_in[4].value = 1e10;
    CHECK_USERDATA_MEMORY_FAILURES(
        userdata_update_stats(values_in, lenof(values_in)));
    CHECK_USERDATA_MEMORY_FAILURES(
        userdata_retrieve_stats(values_out, lenof(values_out)));
    CHECK_DOUBLEEQUAL(values_out[0].value, 100);
    CHECK_DOUBLEEQUAL(values_out[1].value, 100);
    CHECK_DOUBLEEQUAL(values_out[2].value, 1e10);
    CHECK_DOUBLEEQUAL(values_out[3].value, 1e10);
    CHECK_DOUBLEEQUAL(values_out[4].value, 1);
    CHECK_USERDATA_MEMORY_FAILURES(
        userdata_register_stats(stats, lenof(stats)));
    CHECK_USERDATA_MEMORY_FAILURES(
        userdata_retrieve_stats(values_out, lenof(values_out)));
    CHECK_DOUBLEEQUAL(values_out[0].value, 100);
    CHECK_DOUBLEEQUAL(values_out[1].value, 100);
    CHECK_DOUBLEEQUAL(values_out[2].value, 1e10);
    CHECK_DOUBLEEQUAL(values_out[3].value, 1e10);
    CHECK_DOUBLEEQUAL(values_out[4].value, 1);

    CHECK_TRUE(path_exists(PATH_STATS_DIR));
    CHECK_TRUE(path_exists(PATH_STATS_FILE));
    CHECK_TRUE(path_exists(PATH_STATS_DIR "/PARAM.SFO"));
    CHECK_FALSE(path_exists(PATH_STATS_DIR "/ICON0.PNG"));

    remove_or_warn(PATH_STATS_FILE, 0);
    remove_or_warn(PATH_STATS_DIR "/PARAM.SFO", 0);
    remove_or_warn(PATH_STATS_DIR, 1);
    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
